package be.iminds.ilabt.jfed.espec.util;

import be.iminds.ilabt.jfed.espec.model.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.function.Consumer;

/**
 * ESpecLogger helper for managing ESpecLogListeners
 *
 * This ignores or uses an UncaughtExceptionHandler (like UncaughtGuiExceptionToBugReport) to handle exceptions in listeners,
 * instead of propagating them, as they are typically not fatal and not relevant to the method that fires the event.
 */
public class ESpecLogger {
    private static final Logger LOG = LoggerFactory.getLogger(ESpecLogger.class);

    private final List<ESpecLogListener> listeners = new CopyOnWriteArrayList<>();
    private final AtomicInteger exceptionCount = new AtomicInteger(0);

    @Nullable private final Thread.UncaughtExceptionHandler uncaughtExceptionHandler;

    public ESpecLogger() {
        this.uncaughtExceptionHandler = Thread.currentThread().getUncaughtExceptionHandler();
    }
    public ESpecLogger(@Nullable Thread.UncaughtExceptionHandler uncaughtExceptionHandler) {
        this.uncaughtExceptionHandler = uncaughtExceptionHandler;
    }

    public void addESpecLogListener(@Nonnull ESpecLogListener listener) {
        listeners.add(listener);
    }
    public void removeESpecLogListener(@Nonnull ESpecLogListener listener) {
        listeners.remove(listener);
    }

    public void firePreFileLoad() { executeSafeForAllListeners(l -> l.onPreFileLoad()); }

    public void firePostFileLoadSuccess(@Nonnull FileSource fileSource, boolean isFast, long byteSize)  {
        executeSafeForAllListeners(l -> l.onPostFileLoad(fileSource, isFast, byteSize, new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.SUCCESS)));
    }
    public void firePostFileLoadFail(@Nonnull FileSource fileSource, boolean isFast, long byteSize, @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostFileLoad(fileSource, isFast, byteSize, new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.FAILURE, message, throwable)));
    }
    public void firePostFileLoadAll(boolean success)  {
        executeSafeForAllListeners(l -> l.onPostFileLoadAll(success));
    }

    public void firePreRSpec()  {
        executeSafeForAllListeners(l -> l.onPreRSpec());
    }
    public void fireRequestRSpecKnown(@Nonnull RspecSpec rspecSpec, @Nonnull String requestRSpec)  {
        executeSafeForAllListeners(l -> l.onRequestRSpecKnown(rspecSpec, requestRSpec));
    }
    /**
     * Call this when one experiment is ready (or permanently failed)
     * (Note: currently, multi rspec is not supported, so firePostRSpec is fired once, just before firePostRspecAll)
     *
     * This includes waiting for the experiment to become ready.
     * This (obviously) does not include waiting for dir upload and execute steps of ExperimentSpecification
     * This may or may not include that extra actions are ready
     *   (sharing lans,
     *     waiting for startup scripts if specified in RSpec,
     *     executing ansible as in RSpec,
     *     testing connectivity)
     * It is however recommended that this is called only when these actions are done.
     * */
    public void firePostRSpecSuccess(@Nonnull RspecSpec rspec)  {
        executeSafeForAllListeners(l -> l.onPostRSpec(rspec, new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.SUCCESS)));
    }
    public void firePostRSpecFail(@Nonnull RspecSpec rspec, @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostRSpec(rspec, new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.FAILURE, message, throwable)));
    }
    /**
     * Call this when all experiments are ready (or permanently failed)
     * (Note: currently, multi rspec is not supported, so firePostRSpec is fired once,
     *  just before firePostRspecAll)
     *
     * This includes waiting for the experiment to become ready.
     * This (obviously) does not include waiting for dir upload and execute steps of ExperimentSpecification
     * This may or may not include that extra actions are ready
     *   (sharing lans,
     *     waiting for startup scripts if specified in RSpec,
     *     executing ansible as in RSpec,
     *     testing connectivity)
     * It is however recommended that this is called only when these actions are done.
     * */
    public void firePostRspecAll(boolean success, @Nonnull List<String> allNodeClientIds)  {
        executeSafeForAllListeners(l -> l.onPostRspecAll(success, allNodeClientIds));
    }

    public void firePreDir()  {
        executeSafeForAllListeners(l -> l.onPreDir());
    }
    public void firePreDir(@Nonnull String nodeClientId)  {
        executeSafeForAllListeners(l -> l.onPreDirNode(nodeClientId));
    }
    public void firePreDir(long logEventId, @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId)  {
        executeSafeForAllListeners(l -> l.onPreDir(logEventId, dir, fullPath, nodeClientId));
    }
    public void firePostDirSuccess(long logEventId,
                                   @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId)  {
        executeSafeForAllListeners(l -> l.onPostDir(logEventId, dir, fullPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.SUCCESS)));
    }
    public void firePostDirFail(long logEventId,
                                @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId,
                                @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostDir(logEventId, dir, fullPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.FAILURE, message, throwable)));
    }
    public void firePostDirNode(@Nonnull String nodeClientId, boolean success)  {
        executeSafeForAllListeners(l -> l.onPostDirNode(nodeClientId, success));
    }
    public void firePostDirAll(boolean success)  {
        executeSafeForAllListeners(l -> l.onPostDirAll(success));
    }

    public void firePreUpload()  {
        executeSafeForAllListeners(l -> l.onPreUpload());
    }
    public void firePreUpload(@Nonnull String nodeClientId)  {
        executeSafeForAllListeners(l -> l.onPreUploadNode(nodeClientId));
    }
    public void firePreUpload(long logEventId,
                              @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath,
                              @Nonnull String nodeClientId, UploadProgressTracker uploadProgressTracker)  {
        executeSafeForAllListeners(l -> l.onPreUpload(logEventId, uploadSpec, fullPath, nodeClientId, uploadProgressTracker));
    }
    public void firePostUploadSuccess(long logEventId,
                                      @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath, @Nonnull String nodeClientId)  {
        executeSafeForAllListeners(l -> l.onPostUpload(logEventId, uploadSpec, fullPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.SUCCESS)));
    }
    public void firePostUploadFail(long logEventId,
                                   @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath,
                                   @Nonnull String nodeClientId, @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostUpload(logEventId, uploadSpec, fullPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.FAILURE, message, throwable)));
    }
    public void firePostUploadNode(@Nonnull String nodeClientId, boolean success)  {
        executeSafeForAllListeners(l -> l.onPostUploadNode(nodeClientId, success));
    }
    public void firePostUploadAll(boolean success)  {
        executeSafeForAllListeners(l -> l.onPostUploadAll(success));
    }

    public void firePreExecute()  {
        executeSafeForAllListeners(l -> l.onPreExecute());
    }
    public void firePreExecute(long logEventId,
                               ExecuteSpec executeSpec, String fullPath,
                               String fullLogPath, String nodeClientId,
                               LimitedLiveLog currentLog) {
        executeSafeForAllListeners(l -> l.onPreExecute(
                logEventId,
                executeSpec, fullPath,
                fullLogPath, nodeClientId,
                currentLog));
    }
    public void firePostExecuteSuccess(long logEventId,
                                       @Nonnull ExecuteSpec executeSpec,
                                @Nonnull String fullPath,
                                @Nullable String fullLogPath,
                                @Nonnull String nodeClientId,
                                int exitStatus)  {
        executeSafeForAllListeners(l -> l.onPostExecute(logEventId, executeSpec, fullPath, fullLogPath, nodeClientId, exitStatus,
                new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.SUCCESS)));
    }
    public void firePostExecuteFail(long logEventId,
                                    @Nonnull ExecuteSpec executeSpec,
                                    @Nonnull String fullPath,
                                    @Nullable String fullLogPath,
                                    @Nonnull String nodeClientId,
                                    int exitStatus,
                                    @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostExecute(logEventId, executeSpec, fullPath, fullLogPath, nodeClientId, exitStatus,
                new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.FAILURE, message, throwable)));
    }
    public void firePostExecuteStepAll(@Nonnull ExecuteSpec executeSpec, boolean success)  {
        executeSafeForAllListeners(l -> l.onPostExecuteStepAll(executeSpec, success));
    }
//    public void firePostExecuteNode(@Nonnull String nodeClientId, boolean success)  {
//        executeSafeForAllListeners(l -> l.onPostExecuteNode(nodeClientId, success));
//    }

    public void firePostExecuteAll(boolean success)  {
        executeSafeForAllListeners(l -> l.onPostExecuteAll(success));
    }

    public void firePreAnsible()  {
        executeSafeForAllListeners(l -> l.onPreAnsible());
    }
    public void firePreAnsiblePlaybook(long logEventId,
                                               @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                        @Nonnull String fullPath,
                                        @Nullable String fullLogPath,
                                        @Nonnull String nodeClientId,
                                       @Nonnull LimitedLiveLog limitedLiveLog)  {
        executeSafeForAllListeners(l -> l.onPreAnsiblePlaybook(logEventId, ansiblePlaybookSpec, fullPath, fullLogPath,
                nodeClientId, limitedLiveLog));
    }
    public void firePostAnsiblePlaybookSuccess(long logEventId,
                                               @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                        @Nonnull String fullPath,
                                        @Nullable String fullLogPath,
                                        @Nonnull String nodeClientId)  {
        executeSafeForAllListeners(l -> l.onPostAnsiblePlaybook(logEventId, ansiblePlaybookSpec, fullPath, fullLogPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.SUCCESS)));
    }
    public void firePostAnsiblePlaybookFail(long logEventId,
                                            @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                            @Nonnull String fullPath,
                                            @Nullable String fullLogPath,
                                            @Nonnull String nodeClientId,
                                            @Nullable String message, @Nullable Throwable throwable)  {
        executeSafeForAllListeners(l -> l.onPostAnsiblePlaybook(logEventId, ansiblePlaybookSpec, fullPath, fullLogPath, nodeClientId,
                new ESpecLogListener.ESpecStepStatus(ESpecLogListener.ESpecStepStatus.Status.FAILURE, message, throwable)));
    }
    public void firePostAnsibleAll(boolean success)  {
        executeSafeForAllListeners(l -> l.onPostAnsibleAll(success));
    }

    private void executeSafeForAllListeners(Consumer<ESpecLogListener> c) {
        for (ESpecLogListener l : listeners) {
            try {
                c.accept(l);
            } catch (Exception e) {
                handleExceptionInListener(e);
            }
        }
    }

    /**
     * We only handle the first exception with an uncaughtExceptionhandler, if present.
     * Otherwise, the user might see too much similar exceptions.
     *
     * @param e
     */
    private void handleExceptionInListener(Exception e) {
        int count = exceptionCount.getAndIncrement();
        if (count == 0 && uncaughtExceptionHandler != null) {
            try {
                uncaughtExceptionHandler.uncaughtException(Thread.currentThread(), e);
            } catch (Exception e2) {
                LOG.error("Ignoring exception in uncaughtExceptionHandler", e2);
            }
        } else {
            LOG.error("Ignoring exception "+count+" in ESpecLogger listener", e);
        }
    }


    private final static AtomicLong idGenerator = new AtomicLong(0l);
    public static long getNextLogId() {
        return idGenerator.getAndIncrement();
    }
}
